#define LOG_INIT(queue_size)

#define LOG_CLEANUP()

#define LOG( c, i, f )

#define LOG_FILE_WRITE(FileNumber, Percent)

#define LOG_MANDELBROT(Thread, X, Y, FloatX, FloatY, Iterations) \
    (void) Thread; (void) X; (void) Y; (void) Y; \
    (void) FloatX; (void) FloatY; (void) Iterations
